from .dataset import Dataset, DatasetError
import os
import numpy as np
import scipy.io
from deap_utils import macOSFileDatgenerator as datGenera


class Deap(Dataset):
    """An NST dataset.

    An NST dataset usually consists of three files that are within a specific
    subdirectory. The implementation follows this structuring, i.e. the user
    needs to pass a base-directory as well as the identifier upon instantiation.

    """

    def __init__(self, base_dir, identifier, **kwargs):
        """Initialize an NST dataset without loading it.

        Args:
            base_dir (str): The path to the base directory in which the NST dataset resides.
            identifier (str): String identifier for the dataset, e.g. ``S1``
            **kwargs: Additional keyword arguments: n_classes (int, default=3): number of classes to fetch.

        """

        super(Deap, self).__init__(**kwargs)

        self.base_dir = base_dir
        self.data_id = identifier
        self.data_dir = os.path.join(self.base_dir, self.data_id)
        self.data_type = 'EEG'
        self.data_name = 'DEAP'

        # number of classes in the dataset
        self.n_classes = kwargs.pop('n_classes', 2)

        # parameters of the GrazB dataset
        # length of a trial (in seconds)
        self.trial_len = 60
        # motor imagery appears in interval (in seconds)
        self.mi_interval = [0, 60]
        # idle perior prior to start of signal (in seconds)
        self.trial_offset = 0
        # total length of a trial (in seconds)
        self.trial_total = self.trial_len
        # sampling frequency (in Hz)
        self.expected_freq_s = 150

        #DEAP dataset for a subject
        self.fT = os.path.join(self.data_dir, "{id}.dat".format(id=self.data_id))

        # additional variables to store data as expected by the ABC
        self.raw_data = None
        self.trials = None
        self.labels = None
        self.sampling_freq = None

        for f in [self.fT]:
            if not os.path.isfile(f):
                raise DatasetError("Deap Dataset ({id}) file '{f}' unavailable".format(id=self.data_id, f=f))

    def load(self, **kwargs):
        """Loads an NST dataset.

        For more information about the returned values, see
        :meth:`gumpy.data.Dataset.load`
        """
        X = np.array(datGenera.pickle_load(self.fT)['data'])
        X = X[:, :32, :]

        y = np.array(datGenera.pickle_load(self.fT)['labels'])

        y_valence = y[:, 0]
        y_valence = (y_valence > 5).astype(np.int_)

        y_arousal = y[:, 1]
        y_arousal = (y_arousal > 5).astype(np.int_)

        trials_bt = y_valence
        labels_bt = y_valence

        self.raw_data = X[:, :32, :].flatten()
        self.trials = trials_bt
        self.labels = labels_bt
        self.sampling_freq = self.expected_freq_s
        return self
